package org.jvnet.hifaces.myfaces.custom.tree2;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.swing.tree.MutableTreeNode;

import org.apache.commons.lang.ArrayUtils;
import org.apache.myfaces.custom.tree2.TreeNode;
import org.apache.myfaces.util.IteratorEnumeration;

public class SmartTreeNode implements TreeNode, MutableTreeNode {

	private static final long serialVersionUID = 1L;

	private SmartTreeNode parent;

	private String type;

	private String description;

	private final String identifier;

	private boolean expanded;

	private final Object mutex = new Object();

	private final List children = new ArrayList();

	private Map childrenMap = new HashMap();

	private Map descendantsMap = new HashMap();

	private final List unmodifiableChildren = Collections
			.unmodifiableList(children);

	public SmartTreeNode(String identifier, String type, String description) {
		this.identifier = identifier;
		this.type = type;
		this.description = description;
	}

	public javax.swing.tree.TreeNode getParent() {
		return parent;
	}
	
	public SmartTreeNode getSmartParent() {
		return parent;
	}
	

	public String getType() {
		return type;
	}

	public void setType(String type) {
		this.type = type;
	}

	public String getIdentifier() {
		return identifier;
	}

	public void setIdentifier(String identifier) {
		throw new UnsupportedOperationException();
	}

	public String getDescription() {
		return description;
	}

	public void setDescription(String description) {
		this.description = description;
	}

	public boolean isLeaf() {
		return getChildCount() == 0;
	}

	public void setLeaf(boolean leaf) {
		throw new UnsupportedOperationException();
	}

	public int getChildCount() {
		return children.size();
	}

	public List getChildren() {
		return unmodifiableChildren;
	}

	public void addDescendant(SmartTreeNode node) {
		synchronized (mutex) {
			descendantsMap.put(node.getIdentifier(), node);
			descendantsMap.putAll(node.getDescendantsMap());
			if (parent != null) {
				parent.addDescendant(node);
			}
		}
	}

	public void removeDescendant(SmartTreeNode node) {
		synchronized (mutex) {
			descendantsMap.remove(node.getIdentifier());
			for (final Iterator iterator = node.getDescendantsMap().keySet()
					.iterator(); iterator.hasNext();) {
				final String identifier = (String) iterator.next();
				// }
				// for (final String identifier :
				// node.getDescendantsMap().keySet()) {
				descendantsMap.remove(identifier);
			}
			if (parent != null) {
				parent.removeDescendant(node);
			}
		}
	}

	public void addChild(int childIndex, SmartTreeNode node) {
		synchronized (mutex) {
			children.add(childIndex, node);
			childrenMap.put(node.getIdentifier(), node);
			addDescendant(node);
			node.setParent(this);
		}
		if (parent != null) {
			parent.addDescendant(node);
		}
	}

	public void removeChild(int index) {
		synchronized (mutex) {
			final SmartTreeNode node = convert(children.remove(index));
			childrenMap.remove(node.getIdentifier());
			node.setParent(null);
			removeDescendant(node);
		}
	}

	public int getIndex() {
		if (getParent() == null) {
			return 0;
		} else {
			return getParent().getIndex(this);
		}
	}

	public int getIndex(SmartTreeNode node) {
		synchronized (mutex) {
			return children.indexOf(node);
		}
	}

	public boolean isExpanded() {
		return !isLeaf() && expanded;
	}

	public void expand() {
		if (isLeaf())
			expanded = false;
		else
			expanded = true;
	}

	public void collapse() {
		expanded = false;
	}

	public void toggle() {
		if (isLeaf() || isExpanded())
			collapse();
		else
			expand();
	}

	public SmartTreeNode getFirstLeaf() {
		if (isLeaf()) {
			return this;
		} else {
			return getFirstChild().getFirstLeaf();
		}
	}

	protected SmartTreeNode convert(Object node)
	{
		return (SmartTreeNode) node;		
	}
	
	protected boolean isCorrectType(Object node)
	{
		return node == null || node instanceof SmartTreeNode;
	}

	public SmartTreeNode getFirstChild() {
		final SmartTreeNode firstNode;
		synchronized (mutex) {
			firstNode = convert(children.get(0));
		}
		return firstNode;
	}

	public Map getChildrenMap() {
		return childrenMap;
	}

	public Map getDescendantsMap() {
		return descendantsMap;
	}

	public SmartTreeNode getNodeById(String nodeId) {
		if (identifier.equals(nodeId)) {
			return this;
		} else {
			final SmartTreeNode node;
			synchronized (mutex) {
				node = convert(descendantsMap.get(nodeId));
			}
			return node;
		}
	}

	public String[] getPathInformation() {
		if (parent == null) {
			return new String[] { identifier };
		} else {
			final String[] parentPathInfo = parent.getPathInformation();
			return (String[]) ArrayUtils.add(parentPathInfo, identifier);
		}
	}

	public boolean isLastChild(SmartTreeNode node) {
		synchronized (mutex) {
			final int size = children.size();
			return size > 0 && children.get(size - 1) == node;
		}
	}

	public boolean isLastChild() {
		if (parent == null) {
			return true;
		} else {
			return parent.isLastChild(this);
		}
	}

	public SmartTreeNode getFollowingNode() {
		if (parent == null) {
			return null;
		} else {
			return parent.getFollowingNode(this);
		}
	}

	public SmartTreeNode getFollowingNode(SmartTreeNode node) {
		synchronized (mutex) {
			final int index = children.indexOf(node);
			final int size = children.size();
			if (index < 0) {
				return null;
			} else if (index < size - 1) {
				return convert(children.get(index + 1));
			} else {
				if (parent == null) {
					return null;
				} else {
					return parent.getFollowingNode(this);
				}
			}
		}
	}

	public SmartTreeNode getNextNode(boolean checkState) {
		if (children.size() > 0 && (!checkState || isExpanded())) {
			return convert(children.get(0));
		} else if (parent == null) {
			return null;
		} else {
			return getFollowingNode();
		}
	}

	public Enumeration children() {
		return new IteratorEnumeration(children.iterator());
	}

	public boolean getAllowsChildren() {
		return true;
	}

	public javax.swing.tree.TreeNode getChildAt(int childIndex) {
		return getSmartChildAt(childIndex);
	}

	public SmartTreeNode getSmartChildAt(int childIndex) {
		return convert(children.get(childIndex));
	}

	public int getIndex(javax.swing.tree.TreeNode node) {
		return children.indexOf(node);
	}

	public void insert(MutableTreeNode child, int childIndex) {

		if (child == null) {
			throw new IllegalArgumentException("New child is null.");
		} else if (!isCorrectType(child)) {
			throw new IllegalArgumentException("Wrond node type.");
		} else if (isNodeAncestor(convert(child))) {
			throw new IllegalArgumentException("New child is an ancestor.");
		}

		final SmartTreeNode newChild = convert(child);

		SmartTreeNode oldParent = newChild.getSmartParent();

		if (oldParent != null) {
			oldParent.remove(newChild);
		}
		newChild.setParent(this);
		addChild(childIndex, newChild);
	}

	public void remove(int childIndex) {
		final SmartTreeNode child = getSmartChildAt(childIndex);
		removeChild(childIndex);
		child.setParent(null);
	}

	public void remove(MutableTreeNode child) {
		if (child == null) {
			throw new IllegalArgumentException("Argument is null.");
		} else if (!isCorrectType(child)) {
			throw new IllegalArgumentException("Wrong node type.");
		}

		final SmartTreeNode aChild = convert(child);

		if (!isNodeChild(aChild)) {
			throw new IllegalArgumentException("argument is not a child");
		}
		remove(getIndex(aChild)); // linear search
	}

	public void removeFromParent() {
		final SmartTreeNode parent = getSmartParent();
		if (parent != null) {
			parent.remove(this);
		}
	}

	public void setParent(MutableTreeNode newParent) {
		if (!(newParent == null || isCorrectType(newParent))) {
			throw new IllegalArgumentException("Wrong node type.");
		}
		parent = convert(newParent);
	}

	public void setUserObject(Object object) {
		throw new UnsupportedOperationException();
	}

	public boolean isNodeAncestor(SmartTreeNode anotherNode) {
		if (anotherNode == null) {
			return false;
		}

		SmartTreeNode ancestor = this;

		do {
			if (ancestor == anotherNode) {
				return true;
			}
		} while ((ancestor = ancestor.getSmartParent()) != null);

		return false;
	}

	public boolean isNodeChild(SmartTreeNode aNode) {
		boolean retval;

		if (aNode == null) {
			retval = false;
		} else {
			if (getChildCount() == 0) {
				retval = false;
			} else {
				retval = (aNode.getParent() == this);
			}
		}
		return retval;
	}

}
